import pytest
import dbus
import xml.etree.ElementTree as ElementTree

INTROSPECTABLE_IFACE = 'org.freedesktop.DBus.Introspectable'

# obj: a dbus proxy object to introspect
# expected_ifaces: sequence of interface names as strings
def check_object_supports_interfaces(obj, expected_ifaces):
    xml_str = str(obj.Introspect(dbus_interface=INTROSPECTABLE_IFACE))
    root = ElementTree.fromstring(xml_str)
    assert root.tag == 'node'

    interface_elements = root.findall('./interface')
    iface_names = map(lambda e: e.attrib['name'], interface_elements)
    iface_names = list(iface_names)
    iface_names.sort()

    expected_ifaces.sort()

    assert iface_names == expected_ifaces

# Test that the root object at /org/a11y/atspi/accessible/root advertises the correct interfaces
def test_introspect_root(registry_root, session_manager):
    check_object_supports_interfaces(
        registry_root,
        ['org.a11y.atspi.Accessible',
         'org.a11y.atspi.Component',
         'org.a11y.atspi.Socket',
         ]
    )

# Test that the registry object at /org/a11y/atspi/registry advertises the correct interfaces
def test_introspect_registry(registry_registry, session_manager):
    check_object_supports_interfaces(
        registry_registry,
        ['org.a11y.atspi.Registry']
    )



'''gnome-session mock template

This module implements mock versions of a few of the DBus interfaces from
org.gnome.SessionManager and org.gnome.SessionManager.ClientPrivate.  They are used by
at-spi2-core's test suite, to simulate communication with a gnome-session manager.
'''

# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 3 of the License, or (at your option) any
# later version.  See http://www.gnu.org/copyleft/lgpl.html for the full text
# of the license.

__author__ = 'Federico Mena Quintero'
__copyright__ = '(c) 2022 Federico Mena Quintero'

import dbus

from dbusmock import MOCK_IFACE
import dbusmock

SYSTEM_BUS = False
BUS_NAME = 'org.gnome.SessionManager'
MAIN_OBJ = '/org/gnome/SessionManager'
MAIN_IFACE = 'org.gnome.SessionManager'
CLIENT_PRIVATE_IFACE = 'org.gnome.SessionManager.ClientPrivate'

def load(mock, parameters):
    mock.is_running = False
    mock.client_serial_number = 0

# FIXME: at-spi2-registryd uses these, but are not implemented here yet:
#   org.gnome.SessionManager iface, SessionOver method
#   org.gnome.SessionManager.ClientPrivate iface, QueryEndSession signal
#   org.gnome.SessionManager.ClientPrivate iface, Stop signala


# org.gnome.SessionManager interface, IsSessionRunning method
@dbus.service.method(MAIN_IFACE, in_signature='', out_signature='b')
def IsSessionRunning(self):
    return self.is_running

# org.gnome.SessionManager interface, RegisterClient method
@dbus.service.method(MAIN_IFACE, in_signature='ss', out_signature='o')
def RegisterClient(self, app_id, client_startup_id):
    client_name = f"client_{self.client_serial_number}"
    self.client_serial_number += 1

    path = '/org/gnome/SessionManager/MockClientPrivate/' + client_name
    self.AddObject(path, CLIENT_PRIVATE_IFACE, {},
                   [
                       ('EndSessionResponse', 'bs', '', ''),
                   ])
    return path

# org.gnome.SessionManager interface, Logout method
@dbus.service.method(MAIN_IFACE, in_signature='u', out_signature='')
def Logout(self, logout_type):
    # only Normal logout is supported in this mock
    assert logout_type == 0

    for path in dbusmock.get_objects():
        if 'MockClientPrivate/' in path:
            obj = dbusmock.get_object(path)
            obj.EmitSignal(CLIENT_PRIVATE_IFACE, 'EndSession', 'u', [ 0 ])
            obj.EmitSignal(CLIENT_PRIVATE_IFACE, 'Stop', '', [])

# org.freedesktop.DBus.Mock interface, SetSessionRunning method
#
# at-spi2-registryd needs to know the difference between the session
# being in the "running" state or not, so this method is a way to be able to change that from
# the test code.
#
# This will change the internal state and emit the SessionRunning signal
# on the main interface, since at-spi2-registryd handles that signal.
@dbus.service.method(MOCK_IFACE, in_signature='b', out_signature='')
def SetSessionRunning(self, is_running):
    self.is_running = is_running
    self.EmitSignal(MAIN_IFACE, "SessionRunning", "", ())



import pytest
import dbus

from utils import check_unknown_property_yields_error

def test_root_property_on_invalid_interface(registry_root, session_manager):
    check_unknown_property_yields_error(registry_root, 'this.is.not.a.supported.interface')



#!/usr/bin/env python3
#
# Takes DBus XML files and writes out a pair of introspection.[ch] files for inclusion
# in C code.

import argparse
import os
import sys
from xml.etree import ElementTree

CTEMPLATE = """
/*
 * This file has been auto-generated from the introspection data available
 * in the at-spi2-core repository. The D-Bus protocol is defined in this
 * repository, which can be found at:
 *
 * https://gitlab.gnome.org/GNOME/at-spi2-core
 *
 * DO NOT EDIT.
 */

%s
"""

HTEMPLATE = """
/*
 * This file has been auto-generated from the introspection data available
 * in the at-spi2-core repository. The D-Bus protocol is defined in this
 * repository, which can be found at:
 *
 * https://gitlab.gnome.org/GNOME/at-spi2-core
 *
 * DO NOT EDIT.
 */

#ifndef SPI_INTROSPECTION_DATA_H_
#define SPI_INTROSPECTION_DATA_H_

%s

#endif /* SPI_INTROSPECTION_DATA_H_ */
"""

DECTEMPLATE = """
extern const char *%s;
"""

DEFTEMPLATE = """
const char *%s =
%s;
"""

def convert_name (name):
    return "spi_" + name.replace (".", "_")

def convert_contents (contents):
    contents = contents.replace ("\"", "\\\"")
    literals = ["\"%s\"" % (line) for line in contents.split ("\n")]
    return "\n".join (literals)

def generate_introspection (inputs, c_output_filename, h_output_filename):
    #Open the output files.
    cfile = open (c_output_filename, "w")
    hfile = open (h_output_filename, "w")

    ccontents = ""
    hcontents = ""

    for input_filename in inputs:
        #Open the XML file and process includes.
        try:
            tree = ElementTree.parse (input_filename)
        except Exception as e:
            raise type(e)(f"Invalid XML while parsing {input_filename}: {str(e)}")

        root = tree.getroot ()

        for itf in root.findall ("interface"):
            #Get and convert the name of the interface.
            name = convert_name (itf.attrib["name"])

            contents = convert_contents (ElementTree.tostring (itf, encoding="unicode"))

            hcontents += DECTEMPLATE % (name)
            ccontents += DEFTEMPLATE % (name, contents)

    cfile.write (CTEMPLATE % (ccontents))
    hfile.write (HTEMPLATE % (hcontents))

    cfile.close ()
    hfile.close ()

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Create a C source file and header file from DBus XML files")
    parser.add_argument('sources', metavar='FILE.XML', nargs='+', help='DBus XML interface file')
    parser.add_argument('--c-output', metavar='OUT.C', required=True, help='Name out output C file')
    parser.add_argument('--h-output', metavar='OUT.H', required=True, help='Name out output H file')
    args = parser.parse_args()

    input_filename = sys.argv[1]
    c_output_filename = sys.argv[2]
    h_output_filename = sys.argv[3]

    generate_introspection (args.sources, args.c_output, args.h_output)



# Pytest will pick up this module automatically when running just "pytest".
#
# Each test_*() function gets passed test fixtures, which are defined
# in conftest.py.  So, a function "def test_foo(bar)" will get a bar()
# fixture created for it.

import pytest
import dbus

from utils import get_property, check_unknown_property_yields_error

APPLICATION_IFACE = 'org.a11y.atspi.Application'

def test_application_iface_properties(registry_root, session_manager):
    values = [
        ('ToolkitName',  'at-spi-registry'),
        ('Version',      '2.0'),
    ]

    for prop_name, expected in values:
        assert get_property(registry_root, APPLICATION_IFACE, prop_name) == expected

def test_unknown_property_yields_error(registry_root, session_manager):
    check_unknown_property_yields_error(registry_root, APPLICATION_IFACE)



import dbus
import dbus.service

PROPERTIES_IFACE = 'org.freedesktop.DBus.Properties'
ACCESSIBLE_IFACE = 'org.a11y.atspi.Accessible'
APPLICATION_IFACE = 'org.a11y.atspi.Application'
SOCKET_IFACE = 'org.a11y.atspi.Socket'

ROOT_PATH = '/org/a11y/atspi/accessible/root'

def accessible_get_property(prop_name):
    if prop_name == 'Name':
        pass # s
    elif prop_name == 'Description':
        pass # s
    elif prop_name == 'Parent':
        pass # (so)
    elif prop_name == 'ChildCount':
        pass # i
    elif prop_name == 'Locale':
        pass # s
    elif prop_name == 'AccessibleId':
        pass # s
    else:
        raise ValueError(f'unknown property name {prop_name}')

def application_get_property(prop_name):
    return None

@dbus.service.method(PROPERTIES_IFACE, in_signature='ss', out_signature='v')
def Get(self, iface_name, prop_name):
    if iface_name == ACCESSIBLE_IFACE:
        return accessible_get_property(prop_name)
    elif iface_name == PROPERTY_IFACE:
        return application_get_property(prop_name)

@dbus.service.method(PROPERTIES_IFACE, in_signature='ssv', out_signature='')
def Set(self, iface_name, prop_name, value):
    if iface_name == APPLICATION_IFACE:
        pass # FIXME

def get_accesssibility_bus_address():
    bus = dbus.SessionBus()
    bus_launcher = bus.get_object('org.a11y.Bus', '/org/a11y/bus')
    return str(bus_launcher.GetAddress(dbus_interface='org.a11y.Bus'))

def get_registry_root(a11y_bus):
    return a11y_bus.get_object('org.a11y.atspi.Registry', ROOT_PATH)

class MyObject(dbus.service.Object):
    def __init__(self, a11y_bus, path):
        dbus.service.Object.__init__(self, a11y_bus, path)

if __name__ == '__main__':
    from dbus.mainloop.glib import DBusGMainLoop
    from gi.repository import GLib

    DBusGMainLoop(set_as_default=True)
    main_loop = GLib.MainLoop()

    a11y_address = get_accesssibility_bus_address()
    a11y_bus = dbus.bus.BusConnection(a11y_address)

    registry = get_registry_root(a11y_bus)
    assert registry is not None

    my_unique_name = a11y_bus.get_unique_name()
    my_root = MyObject(a11y_bus, ROOT_PATH)
    
    (registry_bus_name, registry_root_path) = registry.Embed((my_unique_name, my_root), dbus_interface=SOCKET_IFACE)

    print("registry bus name:", registry_bus_name)
    print("registry_root_path:", registry_root_path)

    # main_loop.run()



import pytest
import dbus

COMPONENT_IFACE = 'org.a11y.atspi.Component'

COORD_TYPE_WINDOW = 1
LAYER_WIDGET = 3

def test_contains(registry_root, session_manager):
    assert registry_root.Contains(0, 0, COORD_TYPE_WINDOW, dbus_interface=COMPONENT_IFACE) == False

def test_get_accessible_at_point(registry_root, session_manager):
    (name, path) = registry_root.GetAccessibleAtPoint(0, 0, COORD_TYPE_WINDOW, dbus_interface=COMPONENT_IFACE)
    assert path == '/org/a11y/atspi/null'

def test_get_extents(registry_root, session_manager):
    assert registry_root.GetExtents(COORD_TYPE_WINDOW, dbus_interface=COMPONENT_IFACE) == (0, 0, 1024, 768)

def test_get_position(registry_root, session_manager):
    assert registry_root.GetPosition(COORD_TYPE_WINDOW, dbus_interface=COMPONENT_IFACE) == (0, 0)

def test_get_size(registry_root, session_manager):
    assert registry_root.GetSize(dbus_interface=COMPONENT_IFACE) == (1024, 768)

def test_get_layer(registry_root, session_manager):
    assert registry_root.GetLayer(dbus_interface=COMPONENT_IFACE) == LAYER_WIDGET

def test_get_mdi_z_order(registry_root, session_manager):
    assert registry_root.GetMDIZOrder(dbus_interface=COMPONENT_IFACE) == 0

def test_grab_focus(registry_root, session_manager):
    assert registry_root.GrabFocus(dbus_interface=COMPONENT_IFACE) == False

def test_get_alpha(registry_root, session_manager):
    assert registry_root.GetAlpha(dbus_interface=COMPONENT_IFACE) == 1.0



import pytest
import dbus

PROPERTIES_IFACE = 'org.freedesktop.DBus.Properties'

def get_property(proxy, iface_name, prop_name):
    return proxy.Get(iface_name, prop_name, dbus_interface=PROPERTIES_IFACE)

def check_unknown_property_yields_error(obj, iface):
    with pytest.raises(dbus.exceptions.DBusException):
        get_property(obj, iface, 'NonexistentProperty')



# This file contains common test fixtures for registryd's tests.
# Using pytest's terminology for a test's steps:
#
# 1. Arrange: create a test fixture; this usually involves a running registry and an accessible app stub.
#
# 2. Act: manipulate the app or the registry.
#
# 3. Assert: ensure that the thing being manipulated is in the expected state.
#
# 4. Cleanup: Terminate the registry and the app.  It is important that the registry exits cleanly,
#    so it can write out its code coverage information.
#
# This module exports the following fixtures:
#
# * main_loop - a GLib.MainLoop integrated with the DBusGMainLoop.
#
# * session_manager - A mock gnome-session to control the lifetime of daemons.  In
#   reality, the fixture assumes that there is a gnome-session mock running (see
#   ci/run-registryd-tests.sh) and just tells that mock to Logout at fixture teardown
#   time, so that all daemons that monitor the session's lifetime will exit at teardown.
#
# * registry - A dbus.proxies.ProxyObject for the registry's root object.  This automatically
#   depends on a session_manager fixture to control its lifetime.

import pytest
import dbus

@pytest.fixture
def main_loop():
    from dbus.mainloop.glib import DBusGMainLoop
    from gi.repository import GLib

    DBusGMainLoop(set_as_default=True)
    loop = GLib.MainLoop()
    return loop

def get_accesssibility_bus_address():
    bus = dbus.SessionBus()
    bus_launcher = bus.get_object('org.a11y.Bus', '/org/a11y/bus')
    return str(bus_launcher.GetAddress(dbus_interface='org.a11y.Bus'))

@pytest.fixture
def session_manager():
    # This assumes that pytest is running in this environment:
    #
    # * A session dbus daemon is running
    #
    # * There is a gnome-session mock running
    #
    # See the ci/run-registryd-tests.sh script to see how that environment is set up.

    import time

    bus = dbus.SessionBus()
    mock_session = bus.get_object('org.gnome.SessionManager', '/org/gnome/SessionManager')

    mock_session.SetSessionRunning(True, dbus_interface='org.freedesktop.DBus.Mock')

    yield mock_session

    # Tell all session clients to terminate
    mock_session.Logout(0, dbus_interface='org.gnome.SessionManager')

    # Wait a bit for the a11y bus launcher to really die
    proxy = bus.get_object('org.freedesktop.DBus', '/org/freedesktop/DBus')
    while True:
        names = proxy.ListNames(dbus_interface='org.freedesktop.DBus')
        if 'org.a11y.Bus' in names:
            time.sleep(1)
        else:
            break

    # Reset mock session back to its starting state
    mock_session.Reset(dbus_interface='org.freedesktop.DBus.Mock')

@pytest.fixture
def registry_root(main_loop, session_manager):
    a11y_address = get_accesssibility_bus_address()
    a11y_bus = dbus.bus.BusConnection(a11y_address)

    return a11y_bus.get_object('org.a11y.atspi.Registry', '/org/a11y/atspi/accessible/root')

@pytest.fixture
def registry_registry(main_loop, session_manager):
    a11y_address = get_accesssibility_bus_address()
    a11y_bus = dbus.bus.BusConnection(a11y_address)

    return a11y_bus.get_object('org.a11y.atspi.Registry', '/org/a11y/atspi/registry')



# Pytest will pick up this module automatically when running just "pytest".
#
# Each test_*() function gets passed test fixtures, which are defined
# in conftest.py.  So, a function "def test_foo(bar)" will get a bar()
# fixture created for it.

import pytest
import dbus

from utils import get_property, check_unknown_property_yields_error

ACCESSIBLE_IFACE = 'org.a11y.atspi.Accessible'

ATSPI_ROLE_DESKTOP_FRAME = 14 # see atspi-constants.h

def test_accessible_iface_properties(registry_root, session_manager):
    values = [
        ('Name',        'main'),
        ('Description', ''),
        ('Parent',      ('', '/org/a11y/atspi/null')),
        ('ChildCount',  0),
    ]

    for prop_name, expected in values:
        assert get_property(registry_root, ACCESSIBLE_IFACE, prop_name) == expected

def test_unknown_property_yields_error(registry_root, session_manager):
    check_unknown_property_yields_error(registry_root, ACCESSIBLE_IFACE)

def test_get_child_at_index_for_empty_registry(registry_root, session_manager):
    # FIXME: Shouldn't this return an error, rather than a null reference?
    (name, path) = registry_root.GetChildAtIndex(0, dbus_interface=ACCESSIBLE_IFACE)
    assert path == '/org/a11y/atspi/null'

def test_get_children_for_empty_registry(registry_root, session_manager):
    assert len(registry_root.GetChildren(dbus_interface=ACCESSIBLE_IFACE)) == 0

def test_root_get_index_in_parent(registry_root, session_manager):
    # The registry root is always index 0
    assert registry_root.GetIndexInParent(dbus_interface=ACCESSIBLE_IFACE) == 0

def test_root_get_relation_set(registry_root, session_manager):
    # The registry root has an empty relation set
    assert len(registry_root.GetRelationSet(dbus_interface=ACCESSIBLE_IFACE)) == 0

def test_root_get_role(registry_root, session_manager):
    # Hardcoded to ATSPI_ROLE_DESKTOP_FRAME
    assert registry_root.GetRole(dbus_interface=ACCESSIBLE_IFACE) == ATSPI_ROLE_DESKTOP_FRAME

def test_root_get_role_name(registry_root, session_manager):
    assert registry_root.GetRoleName(dbus_interface=ACCESSIBLE_IFACE) == "desktop frame"

def test_root_get_localized_role_name(registry_root, session_manager):
    # FIXME: see the corresponding FIXME in registry.c, to actually localize this
    assert registry_root.GetLocalizedRoleName(dbus_interface=ACCESSIBLE_IFACE) == "desktop frame"

def test_root_get_state(registry_root, session_manager):
    assert registry_root.GetState(dbus_interface=ACCESSIBLE_IFACE) == [0, 0]

def test_root_get_attributes(registry_root, session_manager):
    assert len(registry_root.GetAttributes(dbus_interface=ACCESSIBLE_IFACE)) == 0

def test_root_get_application(registry_root, session_manager):
    (name, path) = registry_root.GetApplication(dbus_interface=ACCESSIBLE_IFACE)
    assert path == '/org/a11y/atspi/null'
    
def test_root_get_interfaces(registry_root, session_manager):
    ifaces = registry_root.GetInterfaces(dbus_interface=ACCESSIBLE_IFACE)
    assert ifaces.signature == 's'
    assert 'org.a11y.atspi.Accessible' in ifaces
    assert 'org.a11y.atspi.Application' in ifaces
    assert 'org.a11y.atspi.Component' in ifaces
    assert 'org.a11y.atspi.Socket' in ifaces




# AT-SPI - Assistive Technology Service Provider Interface
#
# Copyright 2025 SUSE LLC.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA.

from ..overrides import override
from ..module import get_introspection_module

Atspi = get_introspection_module('Atspi')

__all__ = []


class Accessible(Atspi.Accessible):

    def __bool__(self):
        return True

    def __eq__(a, b):
        return hash(a) == hash(b)

    def __getitem__(self, i):
        len = self.get_child_count()
        if i < 0:
                i = len(self) + i
        if i < 0 or i >= len:
                raise IndexError
        return self.get_child_at_index(i)

    def __hash__(self):
        if hasattr(self, "_hashval"):
            return self._hashval
        if self.app is not None:
            self._hashval = hash(self.app.bus_name + self.path)
        else:
            self._hashval = super.__hash__(self)
        return self._hashval

    def __len__(self):
        try:
            count = self.get_child_count()
        except:
            return 0
        if count < 0:
            return 0
        return count

    def __ne__(a, b):
        return hash(a) != hash(b)

    def __nonzero__(self):
        return True

Accessible = override(Accessible)
__all__.append('Accessible')
